// ----------------------------------------------------------------------
// Camera control class
// ----------------------------------------------------------------------
// 2011.06.16: Support for YUV
// ----------------------------------------------------------------------
// Added the GetYuvData function
// Vertical flip was set to display input images, but that setting was cleared.
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// include system header
// ----------------------------------------------------------------------
#include	<nn/os.h>
#include	<nn/camera.h>
#include	<nn/y2r.h>
#include	<string.h>

// ----------------------------------------------------------------------
// include original header
// ----------------------------------------------------------------------
#include	"CTR_Define.h"
#include	"CtrlCamera.h"

//------------------------------------------------------------
// The thread that processes camera events
//------------------------------------------------------------
void CameraThreadFunc( CCtrlCamera* pCamera )
{
    pCamera->CameraThreadFuncImpl();
}

// ----------------------------------------------------------------------
// CCtrlCamera::CCtrlCamera
// ----------------------------------------------------------------------
// Content
//   Constructor
//   Performs camera initialization.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
CCtrlCamera::CCtrlCamera()
{
	m_width				= 0;
	m_height			= 0;

    // Flag determining whether camera initialization has completed.
    s_IsCameraInitialized = false;

    // Flag to put the camera thread into the wait state.
    s_IsCameraThreadSleep = false;

    // Flag used to stop the camera thread
    s_IsCameraThreadEnd = false;

    // Flag when converting the frame rate
    s_SwitchFrameRateFlag = false;

    // Flag indicating whether there are any images waiting for Y2R conversion.
    s_ExistYuvImage = false;

}

// ----------------------------------------------------------------------
// CCtrlCamera::~CCtrlCamera
// ----------------------------------------------------------------------
// Content
//   Destructor
//   Deallocates the cameras.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
CCtrlCamera::~CCtrlCamera()
{
}


// ----------------------------------------------------------------------
// CCtrlCamera::InitializeCamera
// ----------------------------------------------------------------------
// Content
//   Performs camera class initialization.
// ----------------------------------------------------------------------
// Arguments
//   eCamType		[ I ] Specifies the camera type.
//   eCamSize		[ I ] Specifies the camera image size.
//   eCamRate		[ I ] Specifies the camera frame rate.
// ----------------------------------------------------------------------
// Return value.
//   Always returns 0.
// ----------------------------------------------------------------------
int CCtrlCamera::InitializeCamera(CameraSelect select, Size size, FrameRate rate)
{
	PRINT_OUT("Init OK\n");

	switch(size)
	{
	case SIZE_VGA:			// SIZE_VGA 640x480 
		m_width  = 640;
		m_height = 480;
		break;

	case SIZE_QVGA:			// SIZE_QVGA 320x240 
		m_width  = 320;
		m_height = 240;
		break;

	case SIZE_QQVGA:		// SIZE_QQVGA 160x120
		m_width  = 160;
		m_height = 120;
		break;

	case SIZE_DS_LCD:
		m_width  = 256;
		m_height = 192;
		break;
	}

    m_Size = size;
    m_FrameRate = rate;
    m_Select = select;

    // Start thread for camera processing
    InitializeCameraThread();


	return 0;
}

// ----------------------------------------------------------------------
// CCtrlCamera::FinalizeCamera
// ----------------------------------------------------------------------
// Content
//   Performs camera destruction.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void CCtrlCamera::FinalizeCamera()
{
    // Finalize thread for camera processing.
    FinalizeCameraThread();

    // Free RGB buffer
}


// ======================================================================================
// Functions for camera thread control
// ======================================================================================
#if 0 // This sample does not make any provisions for Sleep and other functions. 
//------------------------------------------------------------
// Preparation before Sleep Mode/transition (Called from main thread.)
//------------------------------------------------------------
void CCtrlCamera::PrepareTransitionCallback(void)
{
    // If camera thread is not finalized, transition to wait state.
    if (!s_IsCameraThreadEnd)
    {
        // Notify camera thread to transition to wait state (to CameraThreadFunc).
        s_IsCameraThreadSleep = true;

        // Signal so that the camera thread does not block with WaitAny (to CameraThreadFunc)
        s_CameraBufferErrorEvent.Signal();

        // Wait for camera thread to transition to wait state (from CameraThreadFunc).
        s_CameraThreadSleepAckEvent.Wait();
    }
}

//------------------------------------------------------------
// Recover processing after Sleep Mode/transition. (Called from main thread.)
//------------------------------------------------------------
void CCtrlCamera::AfterTransitionCallback(void)
{
    // If camera thread is not finalized, cancel the wait state.
    // If recovering in order to finalize the application, here you finalize the camera thread without canceling.
    if (!s_IsCameraThreadEnd && !TransitionHandler::IsExitRequired())
    {
        // Signal to start the camera thread (to CameraThreadFunc)
        s_CameraThreadAwakeEvent.Signal();
    }
}
#endif
//-----------------------------------------------------
// Start camera thread.
//-----------------------------------------------------
void CCtrlCamera::InitializeCameraThread(void)
{
    s_CameraThreadSleepAckEvent.Initialize(false);
    s_CameraThreadAwakeEvent.Initialize(false);
    s_CameraBufferErrorEvent.Initialize(false); // Use when recovering from wait state.

#if 0 // This sample does not make any provisions for Sleep and other functions. 
    TransitionHandler::SetPrepareSleepCallback(PrepareTransitionCallback);
    TransitionHandler::SetAfterSleepCallback(AfterTransitionCallback);
    TransitionHandler::SetPrepareHomeButtonCallback(PrepareTransitionCallback);
    TransitionHandler::SetAfterHomeButtonCallback(AfterTransitionCallback);
#endif
    // Remember the priority of the main thread so you can change the priority later.
    s_MainThreadPriority = nn::os::Thread::GetCurrentPriority();

    // Create a thread for camera processing.
    //   While initializing, the priority is set lower than the main thread.
    //   After initialization has completed, the priority changes to be higher than the main thread.
    s_CameraThread.StartUsingAutoStack(
        CameraThreadFunc, 
        this, 
        4096,
        s_MainThreadPriority + 3);
}

//-----------------------------------------------------
// Finalization of camera thread
//-----------------------------------------------------
void CCtrlCamera::FinalizeCameraThread(void)
{
    // Destroy the thread for camera processing
    s_IsCameraThreadEnd = true;         // Set the end flag
    s_CameraThreadAwakeEvent.Signal();  // Signal so that it does not stop with the sleep state
    s_CameraBufferErrorEvent.Signal();  // Signal so that the camera thread does not stop with WaitAny
    s_CameraThread.Join();              // Wait for thread to end
    s_CameraThread.Finalize();          // Discard thread

#if 0 // This sample does not make any provisions for Sleep and other functions. 
    TransitionHandler::SetPrepareSleepCallback(NULL);
    TransitionHandler::SetAfterSleepCallback(NULL);
    TransitionHandler::SetPrepareHomeButtonCallback(NULL);
    TransitionHandler::SetAfterHomeButtonCallback(NULL);
#endif
    s_CameraThreadSleepAckEvent.Finalize();
    s_CameraThreadAwakeEvent.Finalize();
}

// ======================================================================================
// Camera/Y2R control functions
// ======================================================================================
//-----------------------------------------------------
// Camera initialization
//-----------------------------------------------------
void CCtrlCamera::SetupCamera()
{
    // Before camera capture starts, the following functions must execute in order.
    // 
    //   (1) nn::camera::Initialize()   Initializes the camera library
    //   (2) nn::camera::Activate()     Starts the camera to be used
    //   (3) nn::camera::SetReceiving() Starts transfer
    //   (4) nn::camera::StartCapture() Starts capture

    nn::Result result;

    // Camera library initialization
    while (true)
    {
        // Exit if the end flag is set.
        if (s_IsCameraThreadEnd)
        {
            return;
        }

        result = nn::camera::Initialize();
        if (result.IsSuccess())
        {
            break;
        }
        else if (result == nn::camera::ResultFatalError())
        {
            // Camera restart process failed
            NN_PANIC("Camera has broken.\n");
        }
        else if (result == nn::camera::ResultUsingOtherProcess())
        {
            // Camera is being used by another process
            NN_PANIC("Camera is using by other process.\n");
        }
        else if (result == nn::camera::ResultAlreadyInitialized())
        {
            // Because initialization was already done, no particular process is performed
            NN_LOG("Camera is already initialized.\n");
            break;
        }
        else if (result == nn::camera::ResultIsSleeping())
        {
            // If there is a request to enter the wait state, block until end.
            if (s_IsCameraThreadSleep)
            {
                s_IsCameraThreadSleep = false;
                // Notify that the camera thread has entered the wait state (to Sleep, Home)
                s_CameraThreadSleepAckEvent.Signal();

                // Wait until the thread recovery signal arrives (from Sleep, Home)
                s_CameraThreadAwakeEvent.Wait();
            }
            else
            {
                // Even if there is no request, block for a period of time in order to create an interval to retry.
                nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
            }
            // Retry
            NN_LOG("ShellClose: Retry camera initialization\n");
        }
    }

    // Initialize camera settings
    while (true)
    {
        // Exit if the end flag is set.
        if (s_IsCameraThreadEnd)
        {
            nn::camera::Finalize();
            return;
        }
        result = InitializeCameraSetting();
        if (result.IsSuccess())
        {
            break;
        }
        else if (result == nn::camera::ResultFatalError())
        {
            NN_PANIC("Camera has broken.\n");
        }
        else if (result == nn::camera::ResultIsSleeping())
        {
            // If there is a request to enter the wait state, block until end.
            if (s_IsCameraThreadSleep)
            {
                s_IsCameraThreadSleep = false;
                // Notify that the camera thread has entered the wait state (to Sleep, Home)
                s_CameraThreadSleepAckEvent.Signal();

                // Wait until the thread recovery signal arrives (from Sleep, Home)
                s_CameraThreadAwakeEvent.Wait();
            }
            else
            {
                // Even if there is no request, block for a period of time in order to create an interval to retry.
                nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
            }
            // Retry
            NN_LOG("ShellClose: Retry camera setting\n");
        }
    }

    InitializeY2r();        // Y2R library-related initialization
    InitializeResource();   // Resource settings

    // Initialization has completed, so priority changes to become higher than the main thread.
    nn::os::Thread::ChangeCurrentPriority(s_MainThreadPriority - 3);

    s_IsCameraInitialized = true;
}

//-----------------------------------------------------
// Camera default setting
//-----------------------------------------------------
nn::Result CCtrlCamera::InitializeCameraSetting(void)
{
    // Set the image size for the camera to capture
    NN_UTIL_RETURN_IF_FAILED(
        nn::camera::SetSize(
            m_Select,                // Camera that is the target for the settings
            m_Size,                  // Resolution of the camera to be set
            nn::camera::CONTEXT_A)); // Context where settings are reflected

    // Frame rate setting
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetFrameRate( m_Select, m_FrameRate));

    // Vertical inversion
    NN_UTIL_RETURN_IF_FAILED(nn::camera::FlipImage( m_Select, FLIP_REVERSE, CONTEXT_BOTH));

    // Set trimming
    //   When trimming is performed, the functions related to the trimming types must be called before StartCapture().
    //   
    //   
    nn::camera::SetTrimming(CAMERA_PORT, false);

    // Set number of bytes to transfer
    //   A number of bytes is set in this demo, but if a number of lines is to be set for the number of bytes to transfer, you can get the maximum number of bytes to send using nn::camera::GetMaxLines(), and then use nn::camera::SetTransferLines() to set the value.
    //   
    //    However, there are conditions for the number of bytes being sent, so read the details in section "6.4.3 Capture Settings" in the Programming Manual System.
    //   
    //   
    s_YuvTransferUnit = nn::camera::GetMaxBytes(
        m_width,     // Width of the image to transfer (width after trimming when trimmed)
        m_height);   // Height of the image to transfer (height after trimming when trimmed)
    nn::camera::SetTransferBytes(
        CAMERA_PORT,        // Targeted port
        s_YuvTransferUnit,  // Number of bytes to transfer
        m_width,            // Width of the image to transfer (width after trimming when trimmed)
        m_height);          // Height of the image to transfer (height after trimming when trimmed)

    // Get the Y2R conversion coefficient suited to the data output by the camera.
    //   This value is used when initializing Y2R.
    //   It is unnecessary if you are not using Y2R.
    NN_UTIL_RETURN_IF_FAILED(nn::camera::GetSuitableY2rStandardCoefficient(&s_Coefficient));

	// Photo mode				0 (No correction)
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetPhotoMode(SELECT_CAMERA, PHOTO_MODE_NORMAL));

	// Auto white balance.	OFF
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetAutoWhiteBalance(SELECT_CAMERA, false));  // Disabled.

	// Auto exposure feature				OFF
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetAutoExposure(SELECT_CAMERA, false));      // Disabled.

	// Exposure					0
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetExposure(SELECT_CAMERA, 0));

	// White balance			2 (WF Light)
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetWhiteBalance(SELECT_CAMERA,WHITE_BALANCE_WHITE_FLUORESCENT_LIGHT)); 

	// Sharpness					0
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetSharpness(SELECT_CAMERA,0));

	// Contrast				4 (Pattern 5)
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetContrast(SELECT_CAMERA,CONTRAST_PATTERN_05));

	// Peripheral light ratio				0 (normal)
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetLensCorrection(SELECT_CAMERA,LENS_CORRECTION_ON_70));

    // Noise filter
    NN_UTIL_RETURN_IF_FAILED(nn::camera::SetNoiseFilter(SELECT_CAMERA, true));        // Enabled

    // Start camera
    //   After starting the camera module with this function, call the StartCapture function to start image capture.
    //   
    NN_UTIL_RETURN_IF_FAILED(nn::camera::Activate(SELECT_CAMERA));

    return nn::ResultSuccess();
}  //SetupCamera()

//------------------------------------------------------------
// Initial settings for the Y2R library
//------------------------------------------------------------
void CCtrlCamera::InitializeY2r(void)
{
    // Initialization of the Y2R library
    if (!nn::y2r::Initialize())
    {
        NN_PANIC("Y2R is using by other process.\n");
    }

    // Stop the library explicitly because it cannot be configured during a conversion.
    nn::y2r::StopConversion();
    while (nn::y2r::IsBusyConversion())
    {
        // Wait until conversion ends
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(100));
    }

    // Event initialization
    s_Y2rEndEvent.Initialize(false);

    // Input format settings
    nn::y2r::SetInputFormat(Y2R_INPUT_FORMAT);

    // Output format settings
    nn::y2r::SetOutputFormat(Y2R_OUTPUT_FORMAT);

    // Set angle of rotation for output data
    nn::y2r::SetRotation(nn::y2r::ROTATION_NONE);   // No rotation

    // Set alpha value for output data
    //   Since alpha is not used with RGB 24bit, initialize explicitly.
    nn::y2r::SetAlpha(0xFF);

    // Set the order of the output buffer data
    nn::y2r::SetBlockAlignment(Y2R_BLOCK_ALIGNMENT);

    // Set whether we will receive a notification when the Y2R conversion/transfer is complete.
    nn::y2r::SetTransferEndInterrupt(true);
    nn::y2r::GetTransferEndEvent(&s_Y2rEndEvent);

    // Set width of 1 line of input data
    nn::y2r::SetInputLineWidth(m_width);

    // Set number of vertical lines of input data
    nn::y2r::SetInputLines(m_height);

    // Set standard conversion coefficients
    //   The camera module installed on the CTR system may change in the future.
    //   For this reason, instead of specifying a specific conversion coefficient, we recommend using the GetSuitableY2rStandardCoefficient function to get the conversion coefficient that matches the camera, and using this to set the value.
    //  
    //   
    nn::y2r::SetStandardCoefficient(s_Coefficient);
}  //InitializeY2r()

//------------------------------------------------------------
// Allocate the necessary resources for camera and Y2R processing
//------------------------------------------------------------
void CCtrlCamera::InitializeResource(void)
{
    // Initialize the critical session
    s_CsYuvSwap.Initialize();

    // Initialize the event to acquire from the camera library
    s_CameraRecvEvent.Initialize(false);
    s_CameraVsyncEvent.Initialize(false);

    // Event for buffer error notification
    //   When camera data transfer fails (when FIFO overflowed)
    //   A buffer error notification is made.
    nn::camera::GetBufferErrorInterruptEvent(&s_CameraBufferErrorEvent, CAMERA_PORT);

    // Event for camera V blank notification
    //   During VBlanks, frame rates are calculated and camera settings are changed.
    //   
    nn::camera::GetVsyncInterruptEvent(&s_CameraVsyncEvent, CAMERA_PORT);

    // Allocate a buffer for transferring camera image data (YUV data).
    s_YuvBufferSize = nn::camera::GetFrameBytes(m_width, m_height);
    for (s32 i = 0; i < YUV_BUFFER_NUM; i++)
    {
        NN_ASSERT(!s_pYuvBuffer[i]);
        s_pYuvBuffer[i] = static_cast<u8*>(OS_MALLOC_EX(s_YuvBufferSize, ALIGNMENT_SIZE));
        std::memset(s_pYuvBuffer[i], 0, s_YuvBufferSize);
    }
    s_YuvBufferIndexCapturing = 0;
    s_YuvBufferIndexLatestCaptured = YUV_BUFFER_NUM - 1;
    s_YuvBufferIndexTransmitting = YUV_BUFFER_NUM - 1;

    // Allocate a buffer for transferring the camera image Y2R conversion result (RGB data).
    s_RgbBufferSize = nn::y2r::GetOutputImageSize(m_width, m_height,
        Y2R_OUTPUT_FORMAT);
    NN_ASSERT(!s_pRgbBuffer);
    s_pRgbBuffer = static_cast<u8*>(OS_MALLOC_EX(s_RgbBufferSize, ALIGNMENT_SIZE));
    std::memset(s_pRgbBuffer, 0, s_RgbBufferSize);
} //InitializeResource()

//------------------------------------------------------------
// Free the necessary resources for camera and Y2R processing
//------------------------------------------------------------
void CCtrlCamera::FinalizeResource(void)
{
    // Frees the heap.
    for (s32 i = 0; i < YUV_BUFFER_NUM; i++)
    {
        NN_ASSERT(s_pYuvBuffer[i]);
        OS_FREE(s_pYuvBuffer[i]);
    }
    NN_ASSERT(s_pRgbBuffer);
    OS_FREE(s_pRgbBuffer);

    // Destroys events obtained from the camera library.
    s_CameraRecvEvent.Finalize();
    s_CameraBufferErrorEvent.Finalize();
    s_CameraVsyncEvent.Finalize();

    // Destroys the critical section
    s_CsYuvSwap.Finalize();
}  //FinalizeResource()



//------------------------------------------------------------
// The thread that processes camera events
//------------------------------------------------------------
void CCtrlCamera::CameraThreadFuncImpl( void )
{
    // Camera initialization
    SetupCamera();

    // Exit if a request to finalize the thread comes during camera initialization.
    if (s_IsCameraThreadEnd)
    {
        return;
    }

    // Start camera for firs time and capture settings
    StartCameraCapture();

    // ----------------------------------------
    // Processing by event
    enum
    {
        EVENT_RECV,     // Transfer from the camera is complete
        EVENT_ERROR,    // Restart due to buffer error or malfunction
        EVENT_VSYNC,    // V-Blank interrupt

        EVENT_MAX
    };
    nn::os::WaitObject* pEvent[EVENT_MAX];

    // Register each event
    pEvent[EVENT_RECV] = &s_CameraRecvEvent;
    pEvent[EVENT_ERROR] = &s_CameraBufferErrorEvent;
    pEvent[EVENT_VSYNC] = &s_CameraVsyncEvent;

    while (1)
    {
        // Standby for one of the events registered above
        //   The call to SetReceiving() clears the event notifying the completion of transfer, so only one of the events for reconfiguring the transfer settings can be executed at a time.
        //   
        //   
        if (s_IsCameraThreadEnd)
        {
            NN_LOG("before WaitAny\n");
        }
        s32 eventType = nn::os::WaitObject::WaitAny(pEvent, EVENT_MAX);
        if (s_IsCameraThreadEnd)
        {
            NN_LOG("after WaitAny\n");
        }

        // ----------------------------------------
        // End the thread
        if (s_IsCameraThreadEnd)
        {
            StopCameraCapture();
            break;
        }
        // ----------------------------------------
        // Thread waits
#if 0
        if (s_IsCameraThreadSleep)
        {
            s_IsCameraThreadSleep = false;

            // Stop camera capture
            StopCameraCapture();

            // Notify that the camera thread has entered the sleep wait state (to Sleep, Home)
            s_CameraThreadSleepAckEvent.Signal();

            // Standby until the thread sleep recovery signal arrives (from Sleep, Home)
            NN_LOG("-Sleep camera thread.\n");
            s_CameraThreadAwakeEvent.Wait();
            NN_LOG("-Awake camera thread.\n");

            // ----------------------------------------
            // Process after sleep recovery
            // Cleared once because a Vsync event may have been signaled
            s_CameraVsyncEvent.ClearSignal();

            // Reset the counter for waiting for auto exposure stabilization
            s_FrameCountForStabilize = 0;

            // Use the buffer error mechanism to resume capture.
            s_CameraBufferErrorEvent.Signal();
            continue;
        }
#endif
        switch (eventType)
        {
        case EVENT_RECV:
            // Complete transfer from camera
            {
                // Set transfer for the next frame's portion
                CameraRecvFunc();

                // Request Y2R conversion execution
                s_ExistYuvImage = true;
            }
            break;

        case EVENT_ERROR:
            // Restart due to buffer error, malfunction. Restart after sleep/transition.
            {
                // Capture stops when there is an error, so reconfigure the transfer settings and resume capture operations.
                // 
                StartCameraCapture();
            }
            break;

        case EVENT_VSYNC:
            // V-Blank interrupt
            {
                //   Regardless of the frame rate, change the camera settings when nothing is being transferred and a V-blank interrupt is generated.
                //   
                CameraVsyncFunc();
            }
            break;

        default:
            {
                NN_LOG("Illegal event\n");
            }
            break;
        } //Switch

    } //while

    {
        // Finalization of CAMERA/Y2R
        //   If you do not perform finalization with the following procedure,
        //   there is a possibility that sound noise will be generated in the HOME Menu.
        nn::y2r::StopConversion();                      // (1) Stop Y2R conversion
        nn::camera::StopCapture(CAMERA_PORT);           // (2) Stop capture
        nn::camera::Activate(nn::camera::SELECT_NONE);  // (3) Set all cameras to standby state
        nn::camera::Finalize();                         // (4) End camera
        nn::y2r::Finalize();                            //     End Y2R
    }
    {
        // Release the resources used for CAMERA/Y2R
        FinalizeResource();
    }
} //CameraThreadFunc()

//------------------------------------------------------------
// Main data transfer processing
//------------------------------------------------------------
void CCtrlCamera::CameraRecvFunc(void)
{
    {
        // Switch the write buffer
        //   Block so that this does not overlap with the process of switching the read buffer (Y2rConversion).
        //   
        nn::os::CriticalSection::ScopedLock sl(s_CsYuvSwap);
        s_YuvBufferIndexLatestCaptured = s_YuvBufferIndexCapturing;
        do
        {
            // Switch the ring buffer index
            if ((++s_YuvBufferIndexCapturing) >= YUV_BUFFER_NUM)
            {
                s_YuvBufferIndexCapturing = 0;
            }
        } while (s_YuvBufferIndexCapturing == s_YuvBufferIndexTransmitting);
    }

    // Set next frame transfer
    nn::camera::SetReceiving(
        &s_CameraRecvEvent,                         // Event signaled when transfer completes
        s_pYuvBuffer[s_YuvBufferIndexCapturing],    // Transfer destination of the loaded image data
        CAMERA_PORT,                                // Target port for loading
        s_YuvBufferSize,                            // Size of one frame's portion (size that for transfer completion)
        s_YuvTransferUnit);                         // Size to be transferred at one time
} //CameraRecvFunc()

void CCtrlCamera::CameraVsyncFunc(void)
{
    //   Regardless of the frame rate, change the camera settings when
    //   nothing is being transferred and a V-blank interrupt is generated.
    //   <TODO>
}

//------------------------------------------------------------
// Start camera capture
//------------------------------------------------------------
void CCtrlCamera::StartCameraCapture(void)
{
    // Clears the buffer, error flag
    nn::camera::ClearBuffer(CAMERA_PORT);

    // Data transfer process
    CameraRecvFunc();

    // Start capture
    nn::camera::StartCapture(CAMERA_PORT);

} // StartCameraCapture()

//------------------------------------------------------------
// Stop camera capture
//------------------------------------------------------------
void CCtrlCamera::StopCameraCapture(void)
{
    // Stop capture
    nn::camera::StopCapture(CAMERA_PORT);

    // Calculate the time (milliseconds) for one frame from the set frame rate
    s32 timeout = 1000 / m_FrameRate;
    s32 cnt = 0;
    while (nn::camera::IsBusy(CAMERA_PORT))
    {
        // Depending on the timing when the system is closed, IsBusy() might be in a
        // state where it is always 'true.' For that reason, transition to sleep.
        // If this state takes place in the implementation, you will not be able to
        // exit the loop and transition to sleep.
        // Here, a timeout is used to avoid an infinite loop.
        // Normally, IsBusy is 'true' for at most 1 frame, so
        // set the timeout time somewhat large at 1 frame.
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(1));
        if (++cnt > timeout)
        {
            NN_LOG("Busy timeout\n");
            break;
        }
    }
    // Clears the buffer, error flag
    nn::camera::ClearBuffer(CAMERA_PORT);
} // StopCameraCapture()

//------------------------------------------------------------
// Transfer and convert to Y2R
//------------------------------------------------------------
bool CCtrlCamera::Y2rConversion(void)
{
    // *** BEGIN WARNING ***
    // Due to a hardware bug, when the camera and Y2R are being used at the same time,
    // there is a possibility that the recovery process that takes place when a camera buffer error occurs
    // could cause Y2R transfers to hang, depending on the timing of that recovery process.
    // In this case, the conversion completion event obtained by nn::y2r::GetTransferEndEvent
    // might never be signaled.
    // For details on when this problem occurs and how to deal with it, see the Function Reference Manual for the Y2R library.
    //
    // In this sample, a timeout is inserted in the wait for the above event,
    // and when a timeout occurs, a retry is performed.
    // *** END WARNING ***

    {
        // Switch the camera read buffer.
        //   Block so that this does not overlap with the process of switching the write buffer (CameraRecvFunc).
        nn::os::CriticalSection::ScopedLock sl(s_CsYuvSwap);

        // Set the buffer that has completed acquisition from the camera as the conversion target
        s_YuvBufferIndexTransmitting = s_YuvBufferIndexLatestCaptured;
    }

    s32 cnt = 0;
    while (true)
    {
        // Transfer settings for the Y2R input image (YUV)
        //   Writes YUV data to Y2R. Transfers are measured in lines of data.
        nn::y2r::SetSendingYuv(
            s_pYuvBuffer[s_YuvBufferIndexTransmitting], // Transfer data
            s_YuvBufferSize,                            // Total transfer size
            nn::camera::GetLineBytes(m_width));  // Transfer size for one time (1 line)

        // Transfer setting for the output image (RGB) from Y2R
        //   Load RGB data from Y2R. For the per-time transfer size, we recommend specifying
        //   a size of 8 lines worth in order to boost performance.
        s16 yuvTransferUnit = m_width * 8 * GetOutputFormatBytes(Y2R_OUTPUT_FORMAT);
        nn::y2r::SetReceiving(
            s_pRgbBuffer,                   // Load destination for the converted RGB data
            s_RgbBufferSize,                // Total transfer size
            yuvTransferUnit);               // The size of a single transfer.

        // Start Y2R conversion
        nn::y2r::StartConversion();

        // Standby for notification of Y2R conversion completion event
        //   Waits for Y2R conversion completion because data transfer is aborted if
        //   the transfer for the next conversion process is performed before conversion of the first image has completed.
        //   To deal with a  bug that causes Y2R to hang, a timeout is
        //   inserted that is longer than the time for the conversion process to complete.
        //   See the y2r function reference for the approximate time it takes for conversion.
        //if (s_Y2rEndEvent.Wait(nn::fnd::TimeSpan::FromMilliSeconds(10)))
        if (s_Y2rEndEvent.Wait(nn::fnd::TimeSpan::FromMilliSeconds(100)))
        {
            // Conversion succeeded
            return true;
        }
        else
        {
            // Conversion failed

            NN_LOG("Y2R may have hung up.\n");

            nn::y2r::StopConversion();  // Force conversion to end

            // Although it is very rare for problems to occur over and over,
            // a process is inserted to explicitly exit the loop.
            if (++cnt >= 2)
            {
                // Destroy the conversion at this time
                return false;
            }
            // Timed out, so retry.
        }
    } //while(true)
} //Y2rConversion()

//------------------------------------------------------------
// Get YUV and RGB data
//------------------------------------------------------------
void	CCtrlCamera::Receive(u8* rgbBuffer)
{
    if (s_ExistYuvImage) {
        s_ExistYuvImage = false;
        // Only when a completion event for transferring from the camera is signaled
        // Performs Y2R conversion
        if (Y2rConversion()) {
            std::memcpy( rgbBuffer, s_pRgbBuffer, s_RgbBufferSize);
        }
    }
}

// ----------------------------------------------------------------------
// SetExpose
// ----------------------------------------------------------------------
// Content
//   Specifies the exposure.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between -5 and 5.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetExporse(int n)
{
	nn::camera::CTR::SetExposure(SELECT_IN1, n);
}

// ----------------------------------------------------------------------
// SetAutoExpose
// ----------------------------------------------------------------------
// Content
//   Sets the auto exposure feature to enable/disable.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies 0 (disable) or 1 (enable).
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetAutoExporse(int n)
{
	bool bSet = (n > 0) ? true: false;
	nn::camera::CTR::SetExposure(SELECT_IN1, bSet);
}

// ----------------------------------------------------------------------
// SetSharpness
// ----------------------------------------------------------------------
// Content
//   Specifies the sharpness.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between -4 and 5.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetSharpness(int n)
{
	nn::camera::CTR::SetSharpness(SELECT_IN1, n);
}

// ----------------------------------------------------------------------
// SetPhotoMode
// ----------------------------------------------------------------------
// Content
//   Sets the photography mode.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 4.
//                    0: No correction
//                    1: Portrait
//                    2: Landscape
//                    3: Night view
//                    4: Text photography (QR)
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetPhotoMode(int n)
{
	PhotoMode mode = PHOTO_MODE_NORMAL;
	switch(n)
	{
	case 0:mode = PHOTO_MODE_NORMAL;	break;
	case 1:mode = PHOTO_MODE_PORTRAIT;	break;
	case 2:mode = PHOTO_MODE_LANDSCAPE;	break;
	case 3:mode = PHOTO_MODE_NIGHTVIEW;	break;
	case 4:mode = PHOTO_MODE_LETTER;	break;
	case 5:mode = PHOTO_MODE_MAX;		break;
	default :return;
	}

	nn::camera::CTR::SetPhotoMode(SELECT_IN1, mode);
}

// ----------------------------------------------------------------------
// GetPhotoModeStr
// ----------------------------------------------------------------------
// Content
//   Gets the string for displaying photography mode.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 4.
//                    0: No correction
//                    1: Portrait
//                    2: Landscape
//                    3: Night view
//                    4: Text photography (QR)
//   str		[ O ] Stores the acquired string.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void GetPhotoModeStr(int n, char* str)
{
	switch(n)
	{
	case PHOTO_MODE_NORMAL:		sprintf(str, "Normal");		break;	// No correction. 
	case PHOTO_MODE_PORTRAIT:	sprintf(str, "Portrait");	break;	// Portrait. 
	case PHOTO_MODE_LANDSCAPE:	sprintf(str, "Landscape");	break;	// Landscape. 
	case PHOTO_MODE_NIGHTVIEW:	sprintf(str, "Night view");	break;	// Night view. 
	case PHOTO_MODE_LETTER:		sprintf(str, "Letter(QR)");	break;	// Text photography (QR code) 
	case PHOTO_MODE_MAX:		sprintf(str, "max");		break;	// Error
	}
}

// ----------------------------------------------------------------------
// SetWhiteBalance
// ----------------------------------------------------------------------
// Content
//   Specifies the white balance.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 5.
//                    0: Auto white balance
//                    1: Tungsten light (incandescent light bulb)
//                    2: White fluorescent lighting
//                    3: Sunlight
//                    4: Cloudy/dusk
//                    5: Shade
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetWhiteBalance(int n)
{
	WhiteBalance mode = WHITE_BALANCE_AUTO;
	switch(n)
	{
	case 0:	mode = WHITE_BALANCE_AUTO;	break;
	case 1:	mode = WHITE_BALANCE_3200K;	break;
	case 2:	mode = WHITE_BALANCE_4150K;	break;
	case 3:	mode = WHITE_BALANCE_5200K;	break;
	case 4:	mode = WHITE_BALANCE_6000K;	break;
	case 5:	mode = WHITE_BALANCE_7000K;	break;
	case 6:	mode = WHITE_BALANCE_MAX;	break;	// Error
	default :return;
	}

	nn::camera::CTR::SetWhiteBalance(SELECT_IN1, mode);
}

// ----------------------------------------------------------------------
// GetWhiteBalanceStr
// ----------------------------------------------------------------------
// Content
//   Gets the string for displaying the white balance.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 5.
//                    0: Auto white balance
//                    1: Tungsten light (incandescent light bulb)
//                    2: White fluorescent lighting
//                    3: Sunlight
//                    4: Cloudy/dusk
//                    5: Shade
//   str		[ O ] Stores the acquired string.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void GetWhiteBalanceStr(int n, char* str)
{
	switch(n)
	{
	case WHITE_BALANCE_AUTO:	sprintf(str, "Normal");			break;	// Auto white balance.
	case WHITE_BALANCE_3200K:	sprintf(str, "Tungsten");		break;	// Tungsten light (incandescent light bulb).
	case WHITE_BALANCE_4150K:	sprintf(str, "WF Light");		break;	// White fluorescent lighting.
	case WHITE_BALANCE_5200K:	sprintf(str, "DayLight");		break;	// Sunlight.
	case WHITE_BALANCE_6000K:	sprintf(str, "Cloudy/Horizon");	break;	// Cloudy/dusk
	case WHITE_BALANCE_7000K:	sprintf(str, "Shade");			break;	// Shade.
	case WHITE_BALANCE_MAX:		sprintf(str, "max  ");			break;	// Error
	}
}

// ----------------------------------------------------------------------
// SetAutoWhiteBalance
// ----------------------------------------------------------------------
// Content
//   Sets the auto white balance feature to enable/disable.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies 0 (disable) or 1 (enable).
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetAutoWhiteBalance(int n)
{
	bool bSet = (n > 0) ? true: false;
	// Set to `Normal` temporarily, as configuration only possible when the current setting value is `Normal` (otherwise the configuration fails).
	// Set to `Normal` temporarily

    nn::camera::CTR::SetWhiteBalance(SELECT_IN1, WHITE_BALANCE_AUTO);

	nn::camera::CTR::SetAutoWhiteBalance(SELECT_IN1, bSet);
}

// ----------------------------------------------------------------------
// SetContrast
// ----------------------------------------------------------------------
// Content
//   Specifies contrast.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 11.
//                    The default is set as 6.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetContrast(int n)
{
	Contrast mode = CONTRAST_PATTERN_06;
	switch(n)
	{
	case 0:		mode = CONTRAST_PATTERN_01;	break;
	case 1:		mode = CONTRAST_PATTERN_02;	break;
	case 2:		mode = CONTRAST_PATTERN_03;	break;
	case 3:		mode = CONTRAST_PATTERN_04;	break;
	case 4:		mode = CONTRAST_PATTERN_05;	break;
	case 5:		mode = CONTRAST_PATTERN_06;	break;	// Default
	case 6:		mode = CONTRAST_PATTERN_07;	break;
	case 7:		mode = CONTRAST_PATTERN_08;	break;
	case 8:		mode = CONTRAST_PATTERN_09;	break;
	case 9:		mode = CONTRAST_PATTERN_10;	break;
	case 10:	mode = CONTRAST_PATTERN_11;	break;
	case 11:	mode = CONTRAST_MAX;		break;	// Error
	default:	return;
	}

	nn::camera::CTR::SetContrast(SELECT_IN1, mode);
}

// ----------------------------------------------------------------------
// GetContrastStr
// ----------------------------------------------------------------------
// Content
//   Gets the string for displaying the contrast.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 11.
//   str		[ O ] Stores the acquired string.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void GetContrastStr(int n, char* str)
{
	switch(n)
	{
	case CONTRAST_PATTERN_01:	sprintf(str, "Pattern 01");	break;	// Pattern 01
	case CONTRAST_PATTERN_02:	sprintf(str, "Pattern 02");	break;	// Pattern 02
	case CONTRAST_PATTERN_03:	sprintf(str, "Pattern 03");	break;	// Pattern 03
	case CONTRAST_PATTERN_04:	sprintf(str, "Pattern 04");	break;	// Pattern 04
	case CONTRAST_PATTERN_05:	sprintf(str, "Pattern 05");	break;	// Pattern 05
	case CONTRAST_PATTERN_06:	sprintf(str, "Pattern 06"); break;	// Pattern 06
	case CONTRAST_PATTERN_07:	sprintf(str, "Pattern 07");	break;	// Pattern 07
	case CONTRAST_PATTERN_08:	sprintf(str, "Pattern 08");	break;	// Pattern 08
	case CONTRAST_PATTERN_09:	sprintf(str, "Pattern 09");	break;	// Pattern 09
	case CONTRAST_PATTERN_10:	sprintf(str, "Pattern 10");	break;	// Pattern 10
	case CONTRAST_PATTERN_11:	sprintf(str, "Pattern 11");	break;	// Pattern 11
	case CONTRAST_MAX:			sprintf(str, "max  ");		break;	// Error
	}
}

// ----------------------------------------------------------------------
// SetLensCorrection
// ----------------------------------------------------------------------
// Content
//   Specifies the peripheral light ratio.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 2.
//                    0: Turns peripheral light ratio settings off
//                    1: Adjust so that the peripheral light ratio is 70
//                    2: Adjust so that the peripheral light ratio is 90
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void SetLensCorrection(int n)
{
	LensCorrection mode = LENS_CORRECTION_OFF;
	switch(n)
	{
	case 0:	mode = LENS_CORRECTION_OFF;		break;	// dark
	case 1:	mode = LENS_CORRECTION_ON_70;	break;	// normal
	case 2:	mode = LENS_CORRECTION_ON_90;	break;	// bright
	case 3:	mode = LENS_CORRECTION_MAX;		break;	// Error
	default:	return;
	}

	nn::camera::CTR::SetLensCorrection(SELECT_IN1, mode);
}

// ----------------------------------------------------------------------
// GetLensCorrectionStr
// ----------------------------------------------------------------------
// Content
//   Gets the string for displaying the peripheral light ratio.
// ----------------------------------------------------------------------
// Arguments
//   n			[ I ] Specifies a value between 0 and 2.
//   str		[ O ] Stores the acquired string.
// ----------------------------------------------------------------------
// Return value.
//   None
// ----------------------------------------------------------------------
void GetLensCorrectionStr(int n, char* str)
{
	switch(n)
	{
	case LENS_CORRECTION_OFF:	sprintf(str, "Dark");	break;	// Set peripheral light ratio to OFF
	case LENS_CORRECTION_ON_70:	sprintf(str, "Normal");	break;	// Adjust so that the peripheral light ratio is 70
	case LENS_CORRECTION_ON_90:	sprintf(str, "Bright");	break;	// Adjust so that the peripheral light ratio is 90
	case LENS_CORRECTION_MAX:	sprintf(str, "max");	break;	// Error
	}
}
