// ----------------------------------------------------------------------
// Graphics control class
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// include system header
// ----------------------------------------------------------------------
#include	<nn.h>
#include	<nn/os.h>
#include	<nn/gx.h>
#include	<nn/math.h>

// ----------------------------------------------------------------------
// include original header
// ----------------------------------------------------------------------
#include	"CtrlGraphics.h"
#include	"CTR_Define.h"

// ----------------------------------------------------------------------
// CCtrlGraphics::CCtrlGraphics
// ----------------------------------------------------------------------
// Content
//   Constructor
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
CCtrlGraphics::CCtrlGraphics()
{
	m_PgID					= 0;
	m_ShID					= 0;
	m_HeapForGx				= NULL;

	m_textureWidth			= 0;
	m_textureHeight			= 0;
	m_textureGLDataBuffer	= NULL;
	m_textureDataBuffer		= NULL;
}

// ----------------------------------------------------------------------
// CCtrlGraphics::~CCtrlGraphics
// ----------------------------------------------------------------------
// Content
//   Destructor
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
CCtrlGraphics::~CCtrlGraphics()
{
	FinalizeGraphics();
}

// ----------------------------------------------------------------------
// CCtrlGraphics::ClearDisplay
// ----------------------------------------------------------------------
// Content
//   Clears the specified display.
// ----------------------------------------------------------------------
// Arguments
//   display		[ I ] Specifies the display.
//                        NN_GX_DISPLAY0 Upper
//                        NN_GX_DISPLAY1 Lower
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::ClearDisplay(const s32 display)
{
	m_RenderSystemDrawing.SetRenderTarget(display);
	m_RenderSystemDrawing.Clear();
}

// ----------------------------------------------------------------------
// CCtrlGraphics::SwapBuffer
// ----------------------------------------------------------------------
// Content
//   Updates the specified display.
//   Rendering first occurs when this function is called.
// ----------------------------------------------------------------------
// Arguments
//   display		[ I ] Specifies the display.
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::SwapBuffer(const s32 display)
{
	m_RenderSystemDrawing.SetRenderTarget(display);
	m_RenderSystemDrawing.SwapBuffers();
}

// ----------------------------------------------------------------------
// CCtrlGraphics::InitializeGraphics
// ----------------------------------------------------------------------
// Content
//   Initializes graphics.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   Always returns 0.
// ----------------------------------------------------------------------
u32 CCtrlGraphics::InitializeGraphics()
{
	m_HeapForGx = (uptr)OS_MALLOC(0x800000);

	/* Initialize display */
	m_RenderSystemDrawing.Initialize(m_HeapForGx, 0x800000);

	m_PgID = glCreateProgram();
	m_ShID = glCreateShader(GL_VERTEX_SHADER);

	// Load shader binary with file system
	const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];

	nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE);
	nn::fs::FileReader shaderReader(L"rom:/shader.shbin");

	size_t size = shaderReader.GetSize();
	void* pShader = OS_MALLOC(size);

	s32 read = shaderReader.Read(pShader, size);
	NN_ASSERT(read == size);
	glShaderBinary(1, &m_ShID, GL_PLATFORM_BINARY_DMP, pShader, read);

	glAttachShader(m_PgID, m_ShID);
	glAttachShader(m_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

	glBindAttribLocation(m_PgID, 0, "aPosition");
	glBindAttribLocation(m_PgID, 1, "aColor");

	glLinkProgram(m_PgID);
	glValidateProgram(m_PgID);
	glUseProgram(m_PgID);

	OS_FREE((void*)pShader);

	m_RenderSystemDrawing.SetClearColor(NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f);
	m_RenderSystemDrawing.SetClearColor(NN_GX_DISPLAY1, 0.f, 0.f, 0.f, 1.0f);

	glClearDepthf(1.f);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glFrontFace(GL_CCW);

	glUniform3i(glGetUniformLocation(m_PgID, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
	glUniform3i(glGetUniformLocation(m_PgID, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

	return 0;
}

// ----------------------------------------------------------------------
// CCtrlGraphics::FinalizeGraphics
// ----------------------------------------------------------------------
// Content
//   Destroys graphics.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::FinalizeGraphics()
{
	m_RenderSystemDrawing.Finalize();

	if (m_HeapForGx != NULL)
	{
		OS_FREE((void*)m_HeapForGx);
		m_HeapForGx = NULL;
	}

	if (m_textureGLDataBuffer != NULL)
	{
		OS_FREE(m_textureGLDataBuffer);
		m_textureGLDataBuffer = NULL;
	}

	if (m_textureDataBuffer != NULL)
	{
		OS_FREE(m_textureDataBuffer);
		m_textureDataBuffer = NULL;
	}
}

// ----------------------------------------------------------------------
// CCtrlGraphics::SetColor
// ----------------------------------------------------------------------
// Content
//   Specifies the color to render.
// ----------------------------------------------------------------------
// Arguments
//   display		[ I ] Specifies the display.
//   r				[ I ] Specifies the R component for RGB.(1.0 - 0.0)
//   g				[ I ] Specifies the G component for RGB.(1.0 - 0.0)
//   b				[ I ] Specifies the B component for RGB.(1.0 - 0.0)
// ----------------------------------------------------------------------
// Return value
//   Always returns 0.
// ----------------------------------------------------------------------
void CCtrlGraphics::SetColor(const s32 display, GLfloat r, GLfloat g, GLfloat b)
{
	m_RenderSystemDrawing.SetRenderTarget(display);

	m_RenderSystemDrawing.SetColor(r, g, b);
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawLine
// ----------------------------------------------------------------------
// Content
//   Renders a line in the specified position on the specified display.
// ----------------------------------------------------------------------
// Arguments
//   display			[ I ] Specifies the display.
//   sx					[ I ] Specifies the x-coordinate to start rendering.
//   sy					[ I ] Specifies the y-coordinate to start rendering.
//   ex					[ I ] Specifies the x-coordinate to end rendering.
//   ey					[ I ] Specifies the y-coordinate to end rendering.
// ----------------------------------------------------------------------
// Return value
//   Always returns 0.
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawLine(const s32 display, f32 sx, f32 sy, f32 ex, f32 ey)
{
	m_RenderSystemDrawing.SetRenderTarget(display);

	m_RenderSystemDrawing.SetLineWidth(2.0f);

	m_RenderSystemDrawing.DrawLine(sx, sy, ex, ey);
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawRect
// ----------------------------------------------------------------------
// Content
//   Renders a rectangle in the specified position on the specified display.
// ----------------------------------------------------------------------
// Arguments
//   display			[ I ] Specifies the display.
//   sx					[ I ] Specifies the x-coordinate to start rendering.
//   sy					[ I ] Specifies the y-coordinate to start rendering.
//   ex					[ I ] Specifies the x-coordinate to end rendering.
//   ey					[ I ] Specifies the y-coordinate to end rendering.
// ----------------------------------------------------------------------
// Return value
//   Always returns 0.
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawRect(const s32 display, f32 sx, f32 sy, f32 ex, f32 ey)
{
	m_RenderSystemDrawing.SetRenderTarget(display);

	DrawLine(display, sx, sy, ex, sy);
	DrawLine(display, sx, ey, ex, ey);
	DrawLine(display, sx, sy, sx, ey);
	DrawLine(display, ex, sy, ex, ey);
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawFull
// ----------------------------------------------------------------------
// Content
//   Fills in the specified position on the specified display.
// ----------------------------------------------------------------------
// Arguments
//   display			[ I ] Specifies the display.
//   sx					[ I ] Specifies the x-coordinate to start rendering.
//   sy					[ I ] Specifies the y-coordinate to start rendering.
//   ex					[ I ] Specifies the x-coordinate to end rendering.
//   ey					[ I ] Specifies the y-coordinate to end rendering.
// ----------------------------------------------------------------------
// Return value
//   Always returns 0.
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawFull(const s32 display, f32 sx, f32 sy, f32 ex, f32 ey)
{
	m_RenderSystemDrawing.SetRenderTarget(display);
	m_RenderSystemDrawing.FillRectangle(sx, sy, ex - sx + 1, ey - sy + 1);
}

// ----------------------------------------------------------------------
// GetTextureLength
// ----------------------------------------------------------------------
// Content
//   Gets the optimum texture size.
//   Note: Does not work if the texture is not the size obtained with GetTextureLength
//     
// ----------------------------------------------------------------------
// Arguments
//   imageLength		[I/O] Specifies the size of the image specified in the texture.
// ----------------------------------------------------------------------
// Return value
//   Returns the optimum size.
// ----------------------------------------------------------------------
u32 GetTextureLength(const u32& imageLength)
{        
	u32 textureLength = 8;

	// 8, 16, 32, 64, 128, 256, 512, 1024
	for (u32 i = 0; i < 7; i++)
	{
		if ( imageLength > textureLength )
		{
			textureLength *= 2;
		}
		else
		{
			return textureLength;
		}
	}
	return 1024;
}

// ----------------------------------------------------------------------
// CCtrlGraphics::InitTextureBuffer
// ----------------------------------------------------------------------
// Content
//   Performs initializations for the texture buffer.
//   Re-acquires the buffer only when the size differs from the set size.
// ----------------------------------------------------------------------
// Arguments
//   textureWidth		[ I ] Specifies the texture width.
//   textureHeight		[ I ] Specifies the texture height.
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::InitTextureBuffer(u32 textureWidth, u32 textureHeight)
{
	// Compare to current texture size
	if (textureWidth == m_textureWidth && textureHeight == m_textureHeight)
	{
		return;
	}

	// GLData
	if (m_textureGLDataBuffer != NULL)
	{
		OS_FREE(m_textureGLDataBuffer);
	}
	m_textureGLDataBuffer = (u8*)OS_MALLOC(4 * textureWidth * textureHeight);

	// Data
	if (m_textureDataBuffer != NULL)
	{
		OS_FREE(m_textureDataBuffer);
	}
	m_textureDataBuffer = (u8*)OS_MALLOC(4 * textureWidth * textureHeight);

}

// ----------------------------------------------------------------------
// CCtrlGraphics::GetTextureDataFromBmpFileData
// ----------------------------------------------------------------------
// Content
//   Creates texture data from RGB24 data.
//   Note: One side of the texture must be one of the following sizes
//     8, 16, 32, 64, 128, 256, 512, 1024
// ----------------------------------------------------------------------
// Arguments
//   bmpWidth			[ I ] Specifies the image data width.
//   bmpHeight			[ I ] Specifies the image data height.
//   bmpDataBuffer		[ I ] Specifies the image data buffer.
//   textureWidth		[ O ] Stores the height of the acquired texture data.
//   textureHeight		[ O ] Stores the width of the acquired texture data.
//   quality			[ I ] Specifies the image quality.(1,2,4)
// ----------------------------------------------------------------------
// Return value
//   Returns the texture buffer address.
// ----------------------------------------------------------------------
u8* CCtrlGraphics::GetTextureDataFromBmpFileData(const u32& bmpWidth, const u32& bmpHeight, u8* bmpDataBuffer, 
												u32& textureWidth, u32& textureHeight, u32 quality)
{
	int sp				= quality;
	int setBmpWidth		= bmpWidth  / sp;
	int setBmpHeight	= bmpHeight / sp;
	int addY			= sp;
	int addX			= 4 + (sp - 1) * 4;

	// Convert BMP to OpenGL RGB format
	textureWidth	= GetTextureLength(setBmpWidth);
	textureHeight	= GetTextureLength(setBmpHeight);

	// Prepares the texture buffer
	InitTextureBuffer(textureWidth, textureHeight);
	m_textureWidth	= textureWidth;
	m_textureHeight	= textureHeight;

	// NN_LOG("(1)graphic: bmpWidth       = %d\n", bmpWidth);
	// NN_LOG("(2)graphic: bmpHeight      = %d\n", bmpHeight);
	// NN_LOG("(3)graphic: textureWidth   = %d\n", textureWidth); 
	// NN_LOG("(4)graphic: textureHeight  = %d\n", textureHeight); 

	int sy = m_textureHeight - setBmpHeight;
	// Create data from RGB32 data in the GLDataBuffer in BGRA format
	// Shifts up since (0,0) is the lower left.

	for (u32 y = sy; y < m_textureHeight ; y++)
	{
		u8* textureDataPtr	= m_textureGLDataBuffer + (m_textureWidth * (y)) * 3;
		u8* bmpDataPtr		= bmpDataBuffer         + (bmpWidth * (y - sy) * addY) * 4;	// RGB32

		for (u32 x = 0 ; x < setBmpWidth; x++)
		{
			*textureDataPtr++ = *(bmpDataPtr + 3);
			*textureDataPtr++ = *(bmpDataPtr + 2);
			*textureDataPtr++ = *(bmpDataPtr + 1);
			bmpDataPtr += addX;
		}
	}
 
	// Convert OpenGL RGB format to PICA Native RGB format
	bool result = demo::ConvertGLTextureToNative(GL_RGB_NATIVE_DMP, m_textureWidth, m_textureHeight, m_textureGLDataBuffer, m_textureDataBuffer);
	if (result == false)
	{
		//NN_LOG("  Convert to GL_RGB_NATIVE_DMP failed.\n");
		NN_ASSERT(0);
	}

	return m_textureDataBuffer;
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawCamera
// ----------------------------------------------------------------------
// Content
//   Renders texture on display 1.
// ----------------------------------------------------------------------
// Arguments
//   textureDataPtr		[ I ] Specifies the texture image data buffer.
//   textureWidth		[ I ] Specifies the texture image data width.
//   textureHeight		[ I ] Specifies the texture image data height.
//   bmpWidth			[ I ] Specifies the image data width.
//   bmpHeight			[ I ] Specifies the image data height.
//   quality			[ I ] Specifies the image data quality.
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawCamera(u8* textureDataPtr, u32 textureWidth, u32 textureHeight, u32 bmpWidth, u32 bmpHeight, u32 quality)
{
	static GLuint textureId = 0;
	bmpWidth  /= quality;
	bmpHeight /= quality;

	if( textureId != 0 )
	{
		m_RenderSystemDrawing.DeleteTexture(textureId);
	}

	GLenum internalFormat	= GL_RGB_NATIVE_DMP;
	GLenum format			= GL_RGB_NATIVE_DMP;
	GLenum type				= GL_UNSIGNED_BYTE;

	// Create the texture
	m_RenderSystemDrawing.GenerateTexture(GL_TEXTURE_2D,
		internalFormat,
		textureWidth, textureHeight,
		format, type, textureDataPtr,
		textureId);

	m_RenderSystemDrawing.SetRenderTarget(NN_GX_DISPLAY1);
	m_RenderSystemDrawing.SetColor(0.0f, 0.0f, 1.0f, 0.0f);

	f32 windowPositionX = 0.0f;
	f32 windowPositionY = 0.0f;
	f32 rectangleWidth	= nn::gx::CTR::DISPLAY1_HEIGHT  * textureWidth  / bmpWidth;		// DISPLAY1_HEIGHT = 320
	f32 rectangleHeight = nn::gx::CTR::DISPLAY1_WIDTH   * textureHeight / bmpHeight;	// DISPLAY1_WIDTH  = 256

	m_RenderSystemDrawing.FillTexturedRectangle(textureId,
		windowPositionX, windowPositionY,
		rectangleWidth, rectangleHeight,
		textureWidth, textureHeight,
		textureWidth, textureHeight);
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawDateTime
// ----------------------------------------------------------------------
// Content
//   Renders the current date and time on display 0.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawDateTime(void)
{
	// String data for day of the week
	const char WeekStr[][4] = {"SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"};

	m_RenderSystemDrawing.SetRenderTarget(NN_GX_DISPLAY0);
	m_RenderSystemDrawing.SetColor(1.0f, 1.0f, 1.0f, 1.0f);

	DateTime now = DateTime::GetNow();
	m_RenderSystemDrawing.DrawText(288, 0, "%04d.%02d.%02d %s", now.GetYear(), now.GetMonth(), now.GetDay(), WeekStr[now.GetWeek()]);
	m_RenderSystemDrawing.DrawText(326, 8, "%02d:%02d:%02d", now.GetHour(), now.GetMinute(), now.GetSecond());
}

// ----------------------------------------------------------------------
// CCtrlGraphics::DrawTextEx
// ----------------------------------------------------------------------
// Content
//   Renders string on the specified display.
// ----------------------------------------------------------------------
// Arguments
//   display	[ I ] Specifies the display.
//   x			[ I ] Specifies the x-coordinate to start rendering.
//   y			[ I ] Specifies the y-coordinate to start rendering.
//   fmt		[ I ] Specifies the format of the string to be rendered.
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::DrawTextEx(const s32 display, u32 x, u32 y, char* fmt, ...)
{
	// format text process
	char str[128];
	va_list va;

	va_start(va, fmt);
	vsnprintf(str, 128, fmt, va);
	va_end(va);

	m_RenderSystemDrawing.SetRenderTarget(display);

	m_RenderSystemDrawing.DrawText(x, y, "%s", str);
}

// ----------------------------------------------------------------------
// CCtrlGraphics::WaitVsync
// ----------------------------------------------------------------------
// Content
//   Performs Vsync wait.
// ----------------------------------------------------------------------
// Arguments
//   None
// ----------------------------------------------------------------------
// Return value
//   None
// ----------------------------------------------------------------------
void CCtrlGraphics::WaitVsync()
{
	m_RenderSystemDrawing.WaitVsync(NN_GX_DISPLAY_BOTH);
}
